/*******************************************************************************
 * =============================================================================
 * Copyright 2006 The Apache Software Foundation
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 * =============================================================================
 */

package org.apache.muse.ws.wsdl;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.wsdl.Definition;
import javax.wsdl.Input;
import javax.wsdl.Operation;
import javax.wsdl.Output;
import javax.wsdl.Part;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.xml.namespace.QName;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import org.apache.muse.core.Environment;
import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.util.xml.XsdUtils;
import org.apache.muse.ws.addressing.WsaConstants;

/**
 * 
 * WsdlUtils is a collection of utility methods related to WSDL 1.1.
 * 
 * @author Dan Jemiolo (danj)
 * 
 */

public class WsdlUtils
{
    //
    // Used to lookup all exception messages
    //
    private static Messages _MESSAGES = MessagesFactory.get(WsdlUtils.class);

    /**
     * 
     * The WSDL 1.1 namespace URI.
     * 
     */
    public static final String NAMESPACE_URI = "http://schemas.xmlsoap.org/wsdl/";
    
    public static final String PREFIX = "wsdl";

    /**
     * 
     * The "WSDL-SOAP" namespace used for SOAP bindings.
     * 
     */
    public static final String SOAP_URI = NAMESPACE_URI + "soap/";
    
    public static final String SOAP_PREFIX = "wsdl-soap";
    
    public static final QName ADDRESS_QNAME = new QName(SOAP_URI, "address", SOAP_PREFIX);

    public static final QName BINDING_QNAME = new QName(NAMESPACE_URI, "binding", PREFIX);

    public static final QName DEFINITIONS_QNAME = new QName(NAMESPACE_URI, "definitions", PREFIX);

    public static final QName IMPORT_QNAME = new QName(NAMESPACE_URI, "import", PREFIX);

    public static final QName INPUT_QNAME = new QName(NAMESPACE_URI, "input", PREFIX);

    //
    // below are the names of common WSDL tags
    //

    public static final String LOCATION = "location";

    public static final QName MESSAGE_QNAME = new QName(NAMESPACE_URI, "message", PREFIX);

    public static final String NAME = "name";

    public static final QName OPERATION_QNAME = new QName(NAMESPACE_URI, "operation", PREFIX);

    public static final QName OUTPUT_QNAME = new QName(NAMESPACE_URI, "output", PREFIX);

    public static final QName PORT_QNAME = new QName(NAMESPACE_URI, "port", PREFIX);

    public static final QName PORT_TYPE_QNAME = new QName(NAMESPACE_URI, "portType", PREFIX);

    public static final QName SERVICE_QNAME = new QName(NAMESPACE_URI, "service", PREFIX);

    public static final QName TYPES_QNAME = new QName(NAMESPACE_URI, "types", PREFIX);

    public static final String WSDL4J_VERBOSE_FLAG = "javax.wsdl.verbose";
	
	public static final String DEFAULT_SERVICE_NAME = "MyService";	
    	
	public static final String SERVICES = "services";
	
    private static void addNamespacePrefixes(Element nextSchema, Element finalSchema)
    {
        NamedNodeMap attributes = nextSchema.getAttributes();
        
        for (int i = 0; i < attributes.getLength(); i++)
        {
            Node attribute = attributes.item(i);

            if (!finalSchema.hasAttribute(attribute.getNodeName()))
            {
                String nodePrefix = finalSchema.getPrefix();
                String nodeValue = attribute.getLocalName();

                if (nodePrefix == null || !nodePrefix.equals(nodeValue))
                    finalSchema.setAttribute(attribute.getNodeName(), attribute.getNodeValue());
            }
        }
    }

    /**
     * 
     * This is a convenience method that resolves all of the schema imports
     * referenced in a WSDL. It iterates over every schema in the WSDL's
     * <em>types</em> section and processes it via XsdUtils.addImports().
     * 
     * @return The given WSDL, now with the imported schemas added to its
     *         <em>types</em> section.
     * 
     */
    public static Document addSchemaImports(Document wsdl, String wsdlPath, Environment environment, Set namespaces)
    {
        Element root = XmlUtils.getFirstElement(wsdl);
        Element types = XmlUtils.getElement(root, TYPES_QNAME);
        Element[] schemas = XmlUtils.getElements(types, XsdUtils.SCHEMA_QNAME);

        for (int n = 0; n < schemas.length; ++n)
            addSchemaImports(types, schemas[n], wsdlPath, environment, namespaces);

        return wsdl;
    }

    /**
     * 
     * Reads all of the schemas that are imported by the given schema and adds
     * their content to its parent Node. This method is recursive - it will read
     * through the imported schemas and resolve their imports as well. The
     * method tracks the namespaces is resolves so that no duplicates are
     * imported.
     * 
     * @param schemaContainer
     *            The Element that contains the schema being evaluated. Imported
     *            schemas will be appended to this Element.
     * 
     * @param currentSchema
     *            The schema whose imports will be resolved.
     * 
     * @param path
     *            The path of the document that contains the current schema.
     *            This is used to resolve the paths of the imported schemas.
     * 
     * @param environment
     *            The Muse environment is used to load the imported schema
     *            files.
     * 
     * @param namespaces
     *            The set of namespace URIs that have already been processed.
     *            This allows the method to skip those schemas that have already
     *            been processed via other imports and not duplicate XML.
     * 
     */
    private static void addSchemaImports(Element schemaContainer, Element currentSchema, String path, Environment environment, Set namespaces)
    {
        Element[] imports = XmlUtils.getElements(currentSchema, XsdUtils.IMPORT_QNAME);

        Document doc = schemaContainer.getOwnerDocument();

        for (int n = 0; n < imports.length; ++n)
        {
            String location = imports[n].getAttribute(XsdUtils.SCHEMA_LOCATION);

            //
            // This is to guard against imports that do not have a location
            // such as imports in a WSDL types section that reference
            // sibling schemas.
            //            
            if (location.length() == 0)
                continue;

            String nextPath = environment.createRelativePath(path, location);

            Document importDoc = environment.getDocument(nextPath);

            Element nextSchema = XmlUtils.getFirstElement(importDoc);
            String targetNS = nextSchema.getAttribute(XmlUtils.TARGET_NS);

            //
            // if the target namespace hasn't been seen before, we can add it
            //
            if (!namespaces.contains(targetNS))
            {
                nextSchema = (Element)doc.importNode(nextSchema, true);
                schemaContainer.appendChild(nextSchema);

                //
                // record that we found it before reading through the doc
                // for more imports
                //
                namespaces.add(targetNS);

                addSchemaImports(schemaContainer, nextSchema, nextPath, environment, namespaces);
            }
        }
    }

    /**
     * 
     * This is a convenience method that resolves all of the schema includes
     * referenced in a WSDL. It iterates over every schema in the WSDL's
     * <em>types</em> section and processes it via XsdUtils.addIncludes().
     * 
     * @return The given WSDL, now with the included schemas added to its
     *         <em>types</em> section.
     * 
     */
    public static Document addSchemaIncludes(Document wsdl, String wsdlPath, Environment environment, Set namespaces)
    {
        Element root = XmlUtils.getFirstElement(wsdl);
        Element types = XmlUtils.getElement(root, TYPES_QNAME);
        Element[] schemas = XmlUtils.getElements(types, XsdUtils.SCHEMA_QNAME);

        for (int n = 0; n < schemas.length; ++n)
            addSchemaIncludes(schemas[n], schemas[n], wsdlPath, environment, namespaces);

        return wsdl;
    }

    /**
     * 
     * Reads all of the schemas that are included by the given schema and adds
     * their content to it. The method is recursive - it will read through each
     * included schema and resolve its imports and includes. The method tracks
     * the namespaces is resolves so that no duplicates are imported.
     * 
     * @param finalSchema
     *            The original schema whose includes are being resolved. All
     *            included schema content will be added to this node.
     * 
     * @param currentSchema
     *            The schema whose includes and imports are being resolved. In
     *            the initial call to this method, finalSchema and currentSchema
     *            should be the same node. As the method recurses, currentSchema
     *            will be the included schema that is being processed.
     * 
     * @param path
     *            The path of the document that contains the current schema.
     *            This is used to resolve the paths of the imported schemas.
     * 
     * @param environment
     *            The Muse environment is used to load the included schema
     *            files.
     * 
     * @param namespaces
     *            The set of namespace URIs that have already been processed.
     *            This allows the method to skip those schemas that have already
     *            been processed via other imports/includes and not duplicate
     *            XML.
     * 
     */
    private static void addSchemaIncludes(Element finalSchema, Element currentSchema, String path, Environment environment, Set namespaces)
    {
        Element[] includes = XmlUtils.getElements(currentSchema, XsdUtils.INCLUDE_QNAME);

        for (int n = 0; n < includes.length; ++n)
        {
            String relative = includes[n].getAttribute(XsdUtils.SCHEMA_LOCATION);
            String nextPath = environment.createRelativePath(path, relative);
            
            //
            // copy the included schema content into this schema directly
            //
            Document includeDoc = environment.getDocument(nextPath);
            Element nextSchema = XmlUtils.getFirstElement(includeDoc);
            XmlUtils.moveSubTree(nextSchema, finalSchema, includes[n]);

            //
            // copy namespace prefixes from the schema
            //
            addNamespacePrefixes(nextSchema, finalSchema);

            //
            // now resolve the imports in the included schema - any imports
            // will be added to the schema's parent element (since they have
            // different namespaces)
            //
            Element types = (Element)finalSchema.getParentNode();
            addSchemaImports(types, nextSchema, nextPath, environment, namespaces);

            //
            // finally, recurse and process the new schema's includes
            //
            addSchemaIncludes(finalSchema, nextSchema, nextPath, environment, namespaces);
        }
    }

    public static Definition createDefinition(Document wsdlDoc, File wsdlDir)
    {
        //
        // use WSDL4J to parse/validate the WSDL
        //
        try
        {
            WSDLFactory factory = WSDLFactory.newInstance();
            WSDLReader reader = factory.newWSDLReader();

            //
            // shut off verbose mode in WSDL4J - in a "real world"
            // WSDL with many imports, this generates pages of output
            // that no one can read. we will still get proper error
            // messages if a given XSD/WSDL could not be resolved.
            //
            reader.setFeature(WSDL4J_VERBOSE_FLAG, false);

            return reader.readWSDL(wsdlDir.getAbsolutePath(), wsdlDoc);
        }

        catch (Exception error)
        {
            throw new RuntimeException(error);
        }
    }

    /**
     * 
     * Loads the WSDL at the given path, resolving any imports if desired.
     * 
     * @param environment
     *            The Muse environment is used to load the WSDL file.
     * 
     * @param wsdlPath
     *            The relative path of the WSDL file.
     * 
     * @param resolveReferences
     *            True if you want the method to process and WSDL and XSD
     *            imports that are in the WSDL document. The resulting Document
     *            may be much larger depending on how many imports are
     *            discovered while the method recurses through the other
     *            WSDLs/XSDs.
     * 
     * @return A Document with the WSDL content. If the last parameter was
     *         'true', the Document will have the contents of other WSDLs and
     *         XSDs merged into it.
     * 
     */
    public static Document createWSDL(Environment environment, String wsdlPath, boolean resolveReferences)
    {
        Document wsdl = environment.getDocument(wsdlPath);

        //
        // if they just want the doc, we can quit now
        //
        if (!resolveReferences)
            return wsdl;

        //
        // otherwise we must process all imports/includes, then add in
        // any WSDL imports and merge them together
        //

        Set namespaces = getSchemaNamespaces(wsdl);

        //
        // take care of our imports first...
        //
        addSchemaIncludes(wsdl, wsdlPath, environment, namespaces);
        addSchemaImports(wsdl, wsdlPath, environment, namespaces);

        //
        // ...then create all WSDLs that were imported...
        //
        Collection otherWSDLs = getWsdlImports(wsdl, wsdlPath, environment);
        Iterator i = otherWSDLs.iterator();

        //
        // ...and merge them together
        //
        while (i.hasNext())
        {
            Document next = (Document)i.next();
            mergeWSDL(next, wsdl, namespaces);
        }

        return wsdl;
    }

    public static String getAction(Operation op)
    {
        Input input = op.getInput();
        Object attr = (QName)input.getExtensionAttribute(WsaConstants.ACTION_QNAME);

        if (attr == null)
        {
            Object[] filler = { op.getName(), input.getName(),
                    input.getExtensionAttributes().keySet() };
            throw new RuntimeException(_MESSAGES.get("NoActionOnInput", filler));
        }

        return attr instanceof String ? (String)attr : ((QName)attr).getLocalPart();
    }

    /**
     * 
     * Searches a WSDL document's types section for a complex type defined with 
     * the given name.
     * 
     * @param wsdl
     *            The WDSL document to search.
     * 
     * @param qname
     *            The name of the type definition to find.
     * 
     * @return The schema definition of the given type. The method returns null
     *         if the type is not found.
     * 
     */
    public static Element getComplexTypeDeclaration(Node wsdl, QName qname)
    {
        if (wsdl == null)
            throw new NullPointerException(_MESSAGES.get("NullWSDLDocument"));

        if (qname == null)
            throw new NullPointerException(_MESSAGES.get("NullTypeQName"));

        Element definitions = null;

        if (wsdl.getNodeType() == Node.DOCUMENT_NODE)
            definitions = XmlUtils.getFirstElement(wsdl);

        else
            definitions = (Element)wsdl;
        
        String name = qname.getLocalPart();
        String uri = qname.getNamespaceURI();

        //
        // find all of the xsd:complexType in the wsdl:types section
        //
        Element[] elements = XmlUtils.findInSubTree(definitions, XsdUtils.COMPLEX_TYPE_QNAME);

        //
        // check the @name of each xsd:complexType...
        //
        for (int n = 0; n < elements.length; ++n)
        {
            String nextName = elements[n].getAttribute(XsdUtils.NAME);

            //
            // if we find a match, we still need to check namespaces
            //
            if (nextName.equals(name))
            {
                Element parent = (Element)elements[n].getParentNode();
                String targetNS = parent.getAttribute(XmlUtils.TARGET_NS);

                if (uri.equals(targetNS))
                    return elements[n];
            }
        }

        //
        // type declaration not found
        //
        return null;
    }

    /**
     * 
     * Searches a WSDL document's types section for a type defined with the
     * given name.
     * 
     * @param wsdl
     *            The WDSL document to search.
     * 
     * @param qname
     *            The name of the type definition to find.
     * 
     * @return The schema definition of the given type. The method returns null
     *         if the type is not found.
     * 
     */
    public static Element getElementDeclaration(Node wsdl, QName qname)
    {
        if (wsdl == null)
            throw new NullPointerException(_MESSAGES.get("NullWSDLDocument"));

        if (qname == null)
            throw new NullPointerException(_MESSAGES.get("NullTypeQName"));

        Element definitions = null;

        if (wsdl.getNodeType() == Node.DOCUMENT_NODE)
            definitions = XmlUtils.getFirstElement(wsdl);

        else
            definitions = (Element)wsdl;

        //
        // find all of the xsd:elements in the wsdl:types section
        //
        Element[] elements = XmlUtils.findInSubTree(definitions, XsdUtils.ELEMENT_QNAME);

        String localName = qname.getLocalPart();
        String uri = qname.getNamespaceURI();

        if (uri == null)
        {
            Object[] filler = { qname };
            String message = _MESSAGES.get("NullTypeNamespaceURI", filler);
            throw new RuntimeException(message);
        }

        //
        // check the @name of each xsd:element...
        //
        for (int n = 0; n < elements.length; ++n)
        {
            String nextName = elements[n].getAttribute(XsdUtils.NAME);

            //
            // if we find a match, we still need to check namespaces
            //
            if (nextName.equals(localName))
            {
                Element parent = (Element)elements[n].getParentNode();
                String targetNS = parent.getAttribute(XmlUtils.TARGET_NS);

                if (uri.equals(targetNS))
                    return elements[n];
            }
        }

        //
        // type declaration not found
        //
        return null;
    }

    public static QName getInputPartName(Operation op)
    {
        Input input = op.getInput();
        Map parts = input.getMessage().getParts();

        //
        // one message, one part, one XSD element - that's the system
        //
        if (parts.size() != 1)
        {
            Object[] filler = { op.getName() };
            throw new RuntimeException(_MESSAGES.get("NotDocLiteral", filler));
        }

        Part docLiteralPart = (Part)parts.values().iterator().next();
        QName partName = docLiteralPart.getElementName();

        //
        // HACK: this is for Axis/Axiom - they don't handle the default
        // namespace/prefix well, so we always use a prefix
        //
        String prefix = partName.getPrefix();

        if (prefix == null || prefix.length() == 0)
            partName = new QName(partName.getNamespaceURI(), partName.getLocalPart(), "muse-op");

        return partName;
    }

    /**
     * 
     * This method is somewhat painful because of the lack of restrictions that
     * WSDL 1.1 has for how you can specify input and output names and types.
     * Similar to the problem of generating proxies that have the proper return
     * type/name, determining how to package response messages is hard because
     * there are multiple ways to say the same thing (void, xsd:any, etc.) and
     * almost everything is optional.
     * 
     * This method is way too long, but I don't think breaking up the input and
     * output message processing would help clarify things. The output message
     * processing is most problematic, because each step of the way is optional.
     * This is going to be ugly no matter what.
     * 
     */
    public static QName getOutputPartName(Operation op, Document wsdlDoc)
    {
        Output output = op.getOutput();

        //
        // no output message = void
        //
        if (output == null)
            return null;

        Map parts = output.getMessage().getParts();

        //
        // one message, one part, one XSD element - that's the system
        //
        if (parts.size() != 1)
        {
            Object[] filler = { op.getName() };
            throw new RuntimeException(_MESSAGES.get("NotDocLiteral", filler));
        }

        Part docLiteralPart = (Part)parts.values().iterator().next();
        QName responseType = docLiteralPart.getElementName();

        //
        // no output type - this is a void operation
        //
        if (responseType == null)
            return null;

        Element typeDef = getElementDeclaration(wsdlDoc, responseType);

        //
        // empty output message also = pack simple return value into
        // response element (or void)
        //
        if (typeDef.getChildNodes().getLength() == 0)
            return null;

        Element returnValueElement = XmlUtils.findFirstInSubTree(typeDef, XsdUtils.ELEMENT_QNAME);

        //
        // valid output message that is not a complex type or IS
        // a complex type but has an empty sequence(?) = pack
        // simple return value into response element (or void)
        //
        // having fun yet?
        //
        if (returnValueElement == null)
            return null;

        //
        // complex type for output message - use element in the
        // type's sequence as name of the return value. first try 
        // 'name', and then 'ref'.
        //
        // argh.
        //
        String name = returnValueElement.getAttribute(XsdUtils.NAME);
        
        //
        // schema 'name' is unqualified - have to match it against 
        // the targetNamespace of the schema
        //
        if (name != null && name.length() > 0)
            return XmlUtils.parseSchemaName(name, returnValueElement);
        
        //
        // schema 'ref' is qualified - have to resolve the prefix
        //
        String qname = returnValueElement.getAttribute(XsdUtils.REF);
        return XmlUtils.parseQName(qname, returnValueElement);
    }

    /**
     * 
     * Searches a WSDL document for a portType defined with the given name.
     * 
     * @param wsdl
     *            The WDSL document to search.
     * 
     * @param qname
     *            The name of the portType definition to find.
     * 
     * @return The definition of the given portType. The method returns null if
     *         the portType is not found.
     * 
     */
    public static Element getPortType(Node wsdl, QName qname)
    {
        if (wsdl == null)
            throw new NullPointerException(_MESSAGES.get("NullWSDLDocument"));

        if (qname == null)
            throw new NullPointerException(_MESSAGES.get("NullPortTypeQName"));

        //
        // find all the wsdl:portTypes
        //
        Element root = null;

        if (wsdl.getNodeType() == Node.ELEMENT_NODE)
            root = (Element)wsdl;

        else
            root = XmlUtils.getFirstElement(wsdl);

        Element[] portTypes = XmlUtils.getElements(root, PORT_TYPE_QNAME);

        String localName = qname.getLocalPart();
        String uri = qname.getNamespaceURI();

        if (uri == null)
        {
            Object[] filler = { qname };
            String message = _MESSAGES.get("NullPortTypeNamespaceURI", filler);
            throw new RuntimeException(message);
        }

        String targetNS = root.getAttribute(XmlUtils.TARGET_NS);

        //
        // check the @name of each wsdl:portType...
        //
        for (int n = 0; n < portTypes.length; ++n)
        {
            String next = portTypes[n].getAttribute(XsdUtils.NAME);

            //
            // if we find a match, we still need to check namespaces
            //
            if (next.equals(localName) && uri.equals(targetNS))
                return portTypes[n];
        }

        //
        // portType not found
        //
        return null;
    }

    /**
     * 
     * Gathers the target namespaces for each schema in the WSDL's
     * <em>types</em> section.
     * 
     * @param wsdl
     * 
     * @return The set of schema namespaces declared in the <em>types</em>
     *         section.
     * 
     */
    public static Set getSchemaNamespaces(Document wsdl)
    {
        Element root = XmlUtils.getFirstElement(wsdl);
        Element types = XmlUtils.getElement(root, TYPES_QNAME);
        Element[] schemas = XmlUtils.getElements(types, XsdUtils.SCHEMA_QNAME);

        Set namespaces = new HashSet();

        //
        // each schema should have a targetNamespace attribute
        //
        for (int n = 0; n < schemas.length; ++n)
            namespaces.add(schemas[n].getAttribute(XmlUtils.TARGET_NS));

        return namespaces;
    }

    /**
     * Given a WSDL document try to get the address and parse out the name of
     * the service. If we can't find anything useful just return a default: the
     * value of DEFAULT_SERVICE_NAME.
     * 
     * @param wsdlDefinition
     *            The WSDL definition we're trying to parse
     * @return The name of the service in the address element or the value of
     *         DEFAULT_SERVICE_NAME
     * 
     * @see #DEFAULT_SERVICE_NAME
     */
    public static String getServiceName(Element wsdlDefinition)
    {
        Element address = XmlUtils.findFirstInSubTree(wsdlDefinition, WsdlUtils.ADDRESS_QNAME);
        if (address != null)
        {
            String location = address.getAttribute(WsdlUtils.LOCATION);

            if (location != null && location.length() > 0)
            {
                try
                {
                    URL url = new URL(location);
                    Pattern pattern = Pattern.compile(SERVICES + "/([A-Za-z0-9_.]+)");
                    Matcher matcher = pattern.matcher(url.getPath());
                    if (matcher.find())
                    {
                        String serviceName = matcher.group(1);
                        if (serviceName.length() > 0)
                            return serviceName;
                    }
                }
                catch (MalformedURLException e)
                {
                    //handled by falling through to bottom
                }
            }
        }
        return DEFAULT_SERVICE_NAME;
    }

    /**
     * 
     * Creates the WSDLs that are imported in the given WSDL as independent
     * Documents. For each WSDL import, the Document is created via
     * createWSDL(), and then <em>its</em> imports are resolved. The resulting
     * set of Documents may have content overlap and should be merged into the
     * original WSDL with mergeWSDL().
     * 
     * @param wsdl
     *            The original WSDL whose imports must be resolved.
     * 
     * @param wsdlPath
     *            The relative path of the given WSDL. This is used to resolve
     *            the paths of the imported WSDLs.
     * 
     * @param environment
     *            The Muse environment is used to load the WSDL files.
     * 
     * @return A Collection of Documents - one for each WSDL imported by the
     *         given WSDL. The Collection may include WSDLs that were imported
     *         by <em>those</em> WSDLs (the method is recursive).
     * 
     */
    public static Collection getWsdlImports(Document wsdl, String wsdlPath, Environment environment)
    {
        return getWsdlImports(wsdl, wsdlPath, environment, new ArrayList());
    }

    /**
     * 
     * Creates the WSDLs that are imported in the given WSDL as independent
     * Documents. For each WSDL import, the Document is created via
     * createWSDL(), and then <em>its</em> imports are resolved. The resulting
     * set of Documents may have content overlap and should be merged into the
     * original WSDL with mergeWSDL().
     * 
     * @param wsdl
     *            The original WSDL whose imports must be resolved.
     * 
     * @param wsdlPath
     *            The relative path of the given WSDL. This is used to resolve
     *            the paths of the imported WSDLs.
     * 
     * @param environment
     *            The Muse environment is used to load the WSDL files.
     * 
     * @param otherWSDLs
     *            The Collection of WSDLs that has been created so far. This
     *            will ultimately be the return value once the recursion ends.
     * 
     * @return A Collection of Documents - one for each WSDL imported by the
     *         given WSDL. The Collection may include WSDLs that were imported
     *         by <em>those</em> WSDLs (the method is recursive).
     * 
     */
    private static Collection getWsdlImports(Document wsdl, String wsdlPath, Environment environment, Collection otherWSDLs)
    {
        Element root = XmlUtils.getFirstElement(wsdl);
        Element[] imports = XmlUtils.getElements(root, IMPORT_QNAME);

        for (int n = 0; n < imports.length; ++n)
        {
            String location = imports[n].getAttribute(LOCATION);

            if (location.length() == 0)
                continue;

            String nextPath = environment.createRelativePath(wsdlPath, location);

            Document nextWSDL = createWSDL(environment, nextPath, true);
            otherWSDLs.add(nextWSDL);

            //
            // recurse - process the imports for this WSDL
            //
            getWsdlImports(nextWSDL, nextPath, environment, otherWSDLs);
        }

        return otherWSDLs;
    }

    /**
     * 
     * Determines which schema content the first WSDL has that the second one
     * doesn't and adds that content to the second WSDL. <br>
     * <br>
     * The method addresses the issue of duplicate schema namespaces by using
     * size: if both WSDLs have a schema for a given namespace, whichever schema
     * is the largest will be put in the final (second) WSDL. This is the best
     * heuristic to get all of the required schema content without analyzing
     * every node in the schema content.
     * 
     * @param from
     *            The WSDL that may have unique schema content.
     * 
     * @param to
     *            The WSDL that will receive the first WSDL's unique schema
     *            content.
     * 
     * @param namespaces
     *            The schema namespaces that have been processed by the second
     *            WSDL.
     * 
     */
    public static void mergeWSDL(Document from, Document to, Set namespaces)
    {
        //
        // this is an expensive operation, although not as expensive as doing
        // a real merge - analyzing every node and adding in the diffs. the
        // merge logic isn't perfect because it could drop some schema content
        // if both WSDLs extend a given schema.
        //

        Element fromRoot = XmlUtils.getFirstElement(from);
        Element toRoot = XmlUtils.getFirstElement(to);

        Element fromTypes = XmlUtils.getElement(fromRoot, TYPES_QNAME);
        Element toTypes = XmlUtils.getElement(toRoot, TYPES_QNAME);

        Element[] fromSchemas = XmlUtils.getElements(fromTypes, XsdUtils.SCHEMA_QNAME);

        Element[] toSchemas = XmlUtils.getElements(toTypes, XsdUtils.SCHEMA_QNAME);
        Map toSchemasByNS = new HashMap();

        //
        // to speed things up, let's lookup all the schemas once and put them
        // in a table
        //
        for (int n = 0; n < toSchemas.length; ++n)
        {
            String targetNS = toSchemas[n].getAttribute(XmlUtils.TARGET_NS);
            toSchemasByNS.put(targetNS, toSchemas[n]);
        }

        for (int n = 0; n < fromSchemas.length; ++n)
        {
            String targetNS = fromSchemas[n].getAttribute(XmlUtils.TARGET_NS);

            //
            // if we've never seen this namespace before, just add all the
            // schema content - done
            //
            if (!namespaces.contains(targetNS))
            {
                fromSchemas[n] = (Element)to.importNode(fromSchemas[n], true);
                toTypes.appendChild(fromSchemas[n]);
            }

            //
            // if we have seen the namespace before, we have to investigate
            //
            else
            {
                Element schema = (Element)toSchemasByNS.get(targetNS);
                Element[] toChildren = XmlUtils.getAllElements(schema);
                Element[] fromChildren = XmlUtils.getAllElements(fromSchemas[n]);

                //
                // NOTE: if the imported WSDL has a larger version of the
                // schema, chances are it includes the same base as the
                // target schema plus extensions, so we replace the target's
                // copy with the import's copy
                //
                // this is kind of a hack, but a full merge would be a real
                // bottleneck, especially since some standard schemas (like
                // WS-A or WS-RP) are included in multiple other standards
                //
                if (fromChildren.length > toChildren.length)
                {
                    toTypes.removeChild(schema);

                    fromSchemas[n] = (Element)to.importNode(fromSchemas[n], true);
                    toTypes.appendChild(fromSchemas[n]);

                    toSchemasByNS.put(targetNS, fromSchemas[n]);
                }
            }

            //
            // either way, make sure we record that we saw this NS
            //
            namespaces.add(targetNS);
        }
    }

    private static Element removeSchemaImports(Element schema)
    {
        //
        // remove <xsd:import> elements
        //
        Element[] imports = XmlUtils.getElements(schema, XsdUtils.IMPORT_QNAME);

        for (int i = 0; i < imports.length; ++i)
            imports[i].removeAttribute(XsdUtils.SCHEMA_LOCATION);

        return schema;
    }

    private static Element removeSchemaIncludes(Element schema)
    {
        //
        // remove <xsd:include> elements
        //
        Element[] includes = XmlUtils.getElements(schema, XsdUtils.INCLUDE_QNAME);

        for (int i = 0; i < includes.length; ++i)
            schema.removeChild(includes[i]);

        return schema;
    }

    public static Element removeSchemaReferences(Element wsdl)
    {
        Element types = XmlUtils.getElement(wsdl, WsdlUtils.TYPES_QNAME);
        Element[] schemas = XmlUtils.getElements(types, XsdUtils.SCHEMA_QNAME);

        for (int i = 0; i < schemas.length; ++i)
        {
            removeSchemaImports(schemas[i]);
            removeSchemaIncludes(schemas[i]);
        }

        return wsdl;
    }

    /**
     * 
     * Iterates over a WSDL document and removes all WSDL imports and XSD
     * imports. This is useful if you are trying to load a remote WSDL into an
     * API such as WSDL4J, which tries to resolve imports. WSDL4J allows you to
     * shut off WSDL imports, but not XSD imports, so removing these before
     * parsing is important.
     * 
     * @param wsdl
     *            The WSDL Document whose import elements will be removed.
     * 
     * @return The given WSDL Document.
     * 
     */
    public static Element removeWsdlReferences(Element wsdl)
    {
        //
        // first remove the <wsdl:import> elements
        //
        Element[] imports = XmlUtils.getElements(wsdl, WsdlUtils.IMPORT_QNAME);

        for (int i = 0; i < imports.length; ++i)
            wsdl.removeChild(imports[i]);

        return wsdl;
    }
}
