/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/


package org.apache.muse.core.platform.osgi;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

import org.apache.muse.core.AbstractEnvironment;
import org.apache.muse.ws.addressing.EndpointReference;
import org.apache.muse.core.platform.osgi.util.OSGiReflectUtilHelper;
import org.osgi.framework.Bundle;

/**
 *
 * OSGiEnvironmentImpl the implementation of the Muse platform <code>Environment</code>
 * for an OSGi deployment
 *
 * @author Joel Hawkins (joelh)
 *
 */

public class OSGiEnvironmentImpl extends AbstractEnvironment implements
		OSGiEnvironment {

	protected ThreadLocal localEnvironmentContext = new ThreadLocal();
	
	/**
	 * constructor for OSGiEnvironmentImpl
	 * @param createSoapClient
	 * @see org.apache.muse.core.AbstractEnvironment#AbstractEnvironment(boolean)
	 */
	public OSGiEnvironmentImpl(boolean createSoapClient) {
		super(createSoapClient);
	}

	/**
	 * method not implemented for the OSGi environment
	 */
	public EndpointReference getDeploymentEPR() {
		return null;
	}

	/**
	 * method not implemented for the OSGi environment
	 */
	public File getRealDirectory() {
		return null;
	}

	private OSGiReflectUtilHelper reflectHelper;

	/**
	 * gets the environment's bundle context for the currently executing thread. 
	 */
	public Bundle getThreadLocalBundle() {
		return (Bundle) localEnvironmentContext.get();
	}

	/**
	 * sets the environment's bundle context for the currently executing thread
	 * @param bundle the <code>Bundle</code> to be used by the currently executing thread
	 */
	public void setThreadLocalBundle(Bundle bundle) {
		localEnvironmentContext.set(bundle);
		if (reflectHelper != null)
			reflectHelper.setThreadLocalBundle(bundle);
	}

	/**
	 * gets an <code>InputStream</code> for a given path relative to the currently
	 * executing Bundle.
	 * 
	 * @param path the path used in the <code>getResource</code> call.
	 * @return the <code>InputStream corresponding to the path, or null if not found.
	 * 
	 * @see org.apache.muse.core.Environment#getDataResourceStream(String) 
	 */
	public InputStream getDataResourceStream(String path) {
		if (path == null)
			throw new NullPointerException("NullResourcePath");

		InputStream input = null;
		Bundle bundle = getThreadLocalBundle();
		if (bundle != null) {
			URL url = bundle.getResource(path);
			if (url != null) {
				try {
					input = url.openStream();
				} catch (IOException ioe) {
					input = null;
				}
			}
		}

		if (input == null)
			input = this.getClass().getResourceAsStream(path);
		if (input == null)
			input = this.getClass().getResourceAsStream("/OSGI-INF/" + path);
		//Last desperate attempt...
		if(input == null){
			input = OSGiReflectUtilHelper.getDefault().getResource(path);
		}
		return input;

	}

}
