/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.core.platform.axis2;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.net.URI;
import java.util.Iterator;
import java.util.Map;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;

import org.w3c.dom.Element;

import org.apache.axiom.om.OMAttribute;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axiom.soap.SOAPHeader;
import org.apache.axis2.context.MessageContext;

import org.apache.muse.core.AbstractEnvironment;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.addressing.EndpointReference;
import org.apache.muse.ws.addressing.MessageHeaders;
import org.apache.muse.ws.addressing.soap.SoapFault;

/**
 *
 * AxisEnvironment is a concrete{@linkplain org.apache.muse.core.Environment Environment} 
 * for the Apache Axis2 SOAP engine. It provides all of the lower-level 
 * functionality that is abstracted through the Environment interface using 
 * the local file system and Axis2 context APIs.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public class AxisEnvironment extends AbstractEnvironment
{
    private static final String _CLASSES = "classes";
    
    //
    // The service installation directory (NOT the WAR file directory). This 
    // is found at $WAR_INSTALL_DIR/WEB-INF/classes
    //
    private File _realDirectory = null;
    
    /**
     * 
     * Converts Axis2's OperationContext into the Muse addressing context, 
     * MessageHeaders. Hopefully this becomes irrelevant through the 
     * adoption of something like JSR-261 or some WS-A project in Apache Commons.
     *
     */
    public MessageHeaders convertContext()
    {
        MessageContext context = MessageContext.getCurrentMessageContext();
        SOAPHeader axiom = context.getEnvelope().getHeader();
        Element dom = convertToDOM(axiom);
        
        try
        {
            return new MessageHeaders(dom);
        }
        
        catch (SoapFault error)
        {
            throw new RuntimeException(error.getMessage(), error);
        }
    }

    /**
     * 
     * Converts Axis2's EPR type into our general one so that code isn't specific 
     * to the Axis2 platform. Hopefully this becomes irrelevant through the 
     * adoption of something like JSR-261 or some WS-A project in Apache Commons.
     *
     */
    public EndpointReference convertEPR(org.apache.axis2.addressing.EndpointReference axisEPR)
    {
        URI address = URI.create(axisEPR.getAddress());
        EndpointReference epr = new EndpointReference(address);
        
        Map parameters = axisEPR.getAllReferenceParameters();
        Iterator i = parameters.keySet().iterator();
        
        while (i.hasNext())
        {
            QName name = (QName)i.next();
            
            OMElement axiomValue = (OMElement)parameters.get(name);
            Element domValue = convertToDOM(axiomValue);
            
            epr.addParameter(domValue);
        }
        
        return epr;
    }
    
    /**
     * 
     * Convert DOM to Axiom. Muse uses the DOM API in the JDK, Axis2 uses 
     * the Axiom API, which is similar but... different.
     *
     */
    public OMElement convertToAxiom(Element xml)
    {
        String xmlString = XmlUtils.toString(xml, false);
        byte[] xmlBytes = xmlString.getBytes();
        StAXOMBuilder builder = null;
        
        try
        {
            builder = new StAXOMBuilder(new ByteArrayInputStream(xmlBytes));
        }
        
        catch (XMLStreamException error)
        {
            throw new RuntimeException(error.getMessage(), error);
        }
        
        return builder.getDocumentElement();
    }
    
    /**
     * 
     * Convert Axiom to DOM. Muse uses the DOM API in the JDK, Axis2 uses 
     * the Axiom API, which is similar but... different.
     *
     */
    public Element convertToDOM(OMElement axiom)
    {
        Element dom = XmlUtils.createElement(axiom.getQName());

        String text = axiom.getText();

        if (text != null && text.length() > 0)
            XmlUtils.setElementText(dom, axiom.getText());

        Iterator i = axiom.getAllAttributes();
        
        while (i.hasNext())
        {
            OMAttribute attr = (OMAttribute)i.next();
            QName qname = attr.getQName();
            String value = attr.getAttributeValue();
            dom.setAttributeNS(qname.getNamespaceURI(), qname.getLocalPart(), value);
        }

        i = axiom.getAllDeclaredNamespaces();
        
        while (i.hasNext())
        {
            OMNamespace ns = (OMNamespace)i.next();
            XmlUtils.setNamespaceAttribute(dom, ns.getPrefix(), ns.getNamespaceURI());
        }

        i = axiom.getChildElements();

        while (i.hasNext())
        {
            Element child = convertToDOM((OMElement)i.next());
            dom.appendChild(child);
        }

        return dom;
    }
    
    /**
     * 
     * This constructor determines the value of the "real directory" - where 
     * the application is installed on the file system so that it can read 
     * local files when it needs to.
     *
     */
    public AxisEnvironment()
    {
        MessageHeaders wsa = convertContext();        
        addAddressingContext(wsa);
        
        MessageContext context = MessageContext.getCurrentMessageContext();
        
        //
        // where are we? this is the path of the service installation, 
        // not of the Axis2 WAR, so we need to go up two levels to 
        // WEB-INF/classes
        //        
        String axisServicePath = context.getAxisService().getFileName().getFile();
        File axisServiceDir = new File(axisServicePath);
        File webInfDir = axisServiceDir.getParentFile().getParentFile();
        _realDirectory = new File(webInfDir, _CLASSES);
        
        String address = context.getTo().getAddress();
        setDefaultURI(getDeploymentURI(address));
    }
    
    public EndpointReference getDeploymentEPR()
    {
        return getDefaultEPR();
    }
    
    public File getRealDirectory()
    {
        return _realDirectory;
    }
}
