/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.tools.generator.projectizer;

import java.io.File;
import java.io.InputStream;
import java.util.Map;

import org.apache.muse.tools.generator.util.ConfigurationData;
import org.apache.muse.tools.generator.util.ConfigurationDataDescriptor;
import org.apache.muse.util.FileUtils;
import org.apache.muse.ws.wsdl.WsdlUtils;
import org.w3c.dom.Document;

/**
 * A <code>Projectizer</code> that creates a project that compiles to
 * an OSGi framework endpoint that can be deployment in a framework like Equinox. 
 * 
 * Handles creating bundle manifests, platform initialization scripts and so on.
 * 
 * @author Andrew Eberbach (aeberbac)
 */
public class OsgiProjectizer extends Axis2Projectizer {

	static ConfigurationDataDescriptor[] REQUIRED_PARAMETERS = 
		new ConfigurationDataDescriptor[] {
			ConfigurationData.FILES_MAP_LIST_CONFIGURATION,
			ConfigurationData.DESCRIPTOR_DOCUMENT_CONFIGURATION,
			ConfigurationData.WSDL_DOCUMENT_LIST_CONFIGURATION, 
			ConfigurationData.OVERWRITE_CONFIGURATION,
			ConfigurationData.CAPABILITIES_MAP_LIST_CONFIGURATION
		};
	
	public void projectize(ConfigurationData data) throws Exception {
		ConfigurationData.checkConfiguration(this, data);
		
		loadParameters(data);									
		
		File templateDir = new File(System.getProperty(Axis2ProjectizerConstants.MUSE_HOME_PROPERTY), OsgiProjectizerConstants.TEMPLATE_DIR);		
		File destDir = new File(_targetDirectory, OsgiProjectizerConstants.PLUGINS_DIR);		
		copyTemplate(templateDir, destDir);

		File modulesDir = new File(System.getProperty(Axis2ProjectizerConstants.MUSE_HOME_PROPERTY),OsgiProjectizerConstants.MODULES_DIR);
		copyJars(OsgiProjectizerConstants.REQUIRED_MODULES, modulesDir, destDir);
		
		File libDir = new File(_targetDirectory, OsgiProjectizerConstants.LIB_DIR);
		copyJars(OsgiProjectizerConstants.REQUIRED_LIBS, modulesDir, libDir);
		
		File descriptorFile = new File(_targetDirectory, OsgiProjectizerConstants.MUSE_DESCRIPTOR_FILE);
				
		File javaSourceDir = new File(_targetDirectory, OsgiProjectizerConstants.JAVA_SRC_DIR);
		createJavaSources(javaSourceDir, _filesMaps);
		
		createBuildFile(_targetDirectory, OsgiProjectizerConstants.BUILD_FILE_RESOURCE, OsgiProjectizerConstants.BUILD_FILE);		
		
		File wsdldir = new File(_targetDirectory, OsgiProjectizerConstants.WSDL_DIR);
		
		createActivatorFile(javaSourceDir, OsgiProjectizerConstants.ACTIVATOR_FILE_RESOURCE);
		
		File routerEntriesDir = new File(_targetDirectory,OsgiProjectizerConstants.ROUTER_ENTRIES_DIR);
		
		createManifest(_targetDirectory, OsgiProjectizerConstants.OSGI_MANIFEST_FILE_RESOURCE, OsgiProjectizerConstants.OSGI_MANIFEST_FILE, getActivatorName());
		
		createConfigIni(_targetDirectory, OsgiProjectizerConstants.CONFIG_FILE_RESOURCE, OsgiProjectizerConstants.CONFIG_FILE);
		
		for(int i=0; i < _capabilitiesList.length; i++) {
			Map capabilities = _capabilitiesList[i];
			Document wsdl = _wsdls[i];
			createDescriptor(_descriptor, wsdl, descriptorFile, capabilities, Axis2ProjectizerConstants.WSDL_RELATIVE_PATH, i);
			createWSDLFile(wsdl, wsdldir);			
			createRouterEntries(routerEntriesDir, WsdlUtils.getServiceName(wsdl.getDocumentElement()), capabilities);
		}
	}
	
	protected void createConfigIni(File destDir, String configFileResource, String configFileName) throws Exception {
		InputStream configIS = FileUtils.loadFromContext(OsgiProjectizer.class,configFileResource );
		
		File configFile = new File(destDir, configFileName);
		copyStreamCheck(configIS, configFile);
	}

	protected String getActivatorName() {
		return getSymbolicName() + "." + OsgiProjectizerConstants.ACTIVATOR_NAME;
	}

	protected void createActivatorFile(File srcDir, String activatorFileResource) throws Exception {
		InputStream activatorTemplateIS = FileUtils.loadFromContext(OsgiProjectizer.class, activatorFileResource);
	
		Object[] filler = { getSymbolicName() };
		String newActivator = loadString(activatorTemplateIS, filler);
		
		String activatorFileName = getActivatorName();
		activatorFileName = activatorFileName.replaceAll("\\.", "\\\\") + ".java";
		
		File activatorFile = new File(srcDir, activatorFileName);
		writeToFileCheck(newActivator, activatorFile);
	}

	protected void createManifest(File baseTargetDir, String manifestFileResource, String manifestFileName, String activatorFileName) throws Exception {
		InputStream manifestTemplateIS = FileUtils.loadFromContext(OsgiProjectizer.class ,manifestFileResource);		
		
		Object[] filler = { getSymbolicName(), activatorFileName, getSymbolicName(), getSymbolicName() };
		String newManifest = loadString(manifestTemplateIS, filler);
		
		File manifestFile = new File(baseTargetDir, manifestFileName);
		writeToFileCheck(newManifest, manifestFile);
	}

	protected Object getSymbolicName() {
		if(_targetDirectory.equals(FileUtils.CURRENT_DIR)) {
			return _targetDirectory.getAbsoluteFile().getParentFile().getName();
		}
		return _targetDirectory.getAbsoluteFile().getName();
	}

	public ConfigurationDataDescriptor[] getConfigurationDataDescriptions() {
		return REQUIRED_PARAMETERS;
	}
}