/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.util.messages;

import java.util.HashMap;
import java.util.Map;

/**
 *
 * MessagesFactory is a static factory that allows any object to look up 
 * a resource bundle associated with its Java package. If the classpath 
 * contains a file named <b>Messages.properties</b> in the same package as 
 * the caller, that properties file will be parsed into a 
 * {@linkplain Messages Messages} collection; this collection can then be 
 * used to lookup external messages.
 * <br><br>
 * This class loads Messages objects only when they are requested and then 
 * caches them for future lookups. It will hold onto the Messages it loads 
 * for the lifetime of the application.
 *
 * @author Dan Jemiolo (danj)
 * 
 * @see Messages
 *
 */

public class MessagesFactory
{
    //
    // Each package can have one Messages.properties resources.
    //
    private static Map _messagesByPackage = new HashMap();
    
    //
    // The standard name for .properties files with error messages.
    //
    private static final String _RESOURCE_NAME = "Messages";
    
    /**
     * 
     * Inspects the object to discover the package name of its Java class 
     * and tries to load a properties file in that package. The file must 
     * be named <b>Messages.properties</b>, reside in the same package (but 
     * not necessarily the same directory or JAR file), and be in the Java 
     * classpath. Once the file is loaded into a Messages, the Messages is 
     * cached for future requests.
     *
     * @param obj
     *        The object that wishes to get the Messages associated with its 
     *        Java package.
     * 
     * @return The Messages collection associated with the appropriate 
     *         Java package.
     *
     */
    public static Messages get(Object obj)
    {
        //
        // parameter could be a class (for static callers) or an instance 
        //
        Class theClass = (obj instanceof Class) ? (Class)obj : obj.getClass();
        
        String packageName = getPackageName(theClass);      
        Messages messages = (Messages)_messagesByPackage.get(packageName);
        
        //
        // bundle hasn't been loaded or does not exist - try to load it 
        //
        if (messages == null)
        {
            String name = _RESOURCE_NAME;
            
            if (packageName != null)
                name = packageName + '.' + _RESOURCE_NAME;
            
            //
            // make sure we use the right classloader to load resources
            //
            messages = new Messages(name, theClass.getClassLoader());
            _messagesByPackage.put(packageName, messages);
        }
        
        return messages;
    }
    
    private static String getPackageName(Class theClass)
    {
        //
        // HACK: This code is copied from ReflectUtils.getPackageName().
        //
        // We can't call ReflectUtils in get() because it has its own 
        // instance of Messages, creating a circular dependency. The 
        // ReflectUtils class must be fully initialized (including the 
        // Messages field) before its methods are invoked.
        //
        
        String name = theClass.getName();
        int dot = name.lastIndexOf('.');
        
        return dot >= 0 ? name.substring(0, dot) : null;
    }
}
