/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.dm.muws.impl;

import java.util.List;

import javax.xml.namespace.QName;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.addressing.EndpointReference;
import org.apache.muse.ws.addressing.soap.SoapFault;
import org.apache.muse.ws.dm.muws.Correlation;
import org.apache.muse.ws.dm.muws.Match;
import org.apache.muse.ws.dm.muws.MuwsConstants;
import org.apache.muse.ws.resource.WsResource;

/**
 *
 * MatchCorrelation represents correlation expressions formatted with the 
 * MUWS Simple Property Boolean Match (PBM) schema. PBM allows the user to 
 * declare fairly complex conditionals for evaluating two resources for 
 * equality. Each MatchCorrelation will contain one or more instances of 
 * {@linkplain Match Match}.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public class MatchCorrelation implements Correlation, Match
{
    //
    // Used to lookup all exception messages
    //
    private static Messages _MESSAGES = MessagesFactory.get(MatchCorrelation.class);
    
    private boolean _hasNegativeAssertion = false;
    
    //
    // A complete Correlation based on PBM is really like a pbm:MatchAll, so 
    // we use that as the container for all the Correlation's Matches. The 
    // only tweak is that we cannot show the pbm:MatchAll root element when 
    // we serialize the Correlation (a CorrelatableProperties instance can 
    // have multiple Matches under its root element).
    //
    private AllMatch _root = null;
    
    /**
     * 
     * This is a convenience constructor that calls this(Resource, boolean) 
     * with a 'false' second parameter.
     * 
     * @see #MatchCorrelation(WsResource, boolean)
     *
     */
    public MatchCorrelation(WsResource resource)
    {
        this(resource, false);
    }
    
    /**
     * 
     * @param resource
     *        The resource that is defining the CorrelatableProperties 
     *        capability and, thus, this correlation.
     * 
     * @param negativeAssertion
     *        True if this correlation guarantees that a false match equates 
     *        to two resources being different.
     *
     */
    public MatchCorrelation(WsResource resource, boolean negativeAssertion)
    {
        if (resource == null)
            throw new NullPointerException(_MESSAGES.get("NullOwner"));
                
        _root = new AllMatch(resource);
        _hasNegativeAssertion = negativeAssertion;
    }
    
    /**
     * 
     * @param resource
     *        The resource that is defining the CorrelatableProperties 
     *        capability and, thus, this correlation.
     * 
     * @param xml
     *        The DOM Element representing an instance of CorrelatableProperties 
     *        that uses the PBM dialect (is defined by Matches).
     *
     */
    public MatchCorrelation(WsResource resource, Element xml)
    {
        if (resource == null)
            throw new NullPointerException(_MESSAGES.get("NullOwner"));
        
        if (xml == null)
            throw new NullPointerException(_MESSAGES.get("NullCPElement"));
        
        _root = new AllMatch(resource, xml);

        //
        // get the (optional) NegativeAssertionPossible value
        //
        String negative = xml.getAttribute(MuwsConstants.NEGATIVE);
        
        if (negative != null && negative.length() > 0)
            _hasNegativeAssertion = Boolean.valueOf(negative).booleanValue();
    }
    
    public void addMatch(Match child)
    {
        if (child == null)
            throw new NullPointerException(_MESSAGES.get("NullMatch"));
        
        _root.addMatch(child);
    }
    
    public String getDialect()
    {
        return MuwsConstants.PBM_URI;
    }

    public List getMatches()
    {
        return _root.getMatches();
    }

    public WsResource getWsResource()
    {
        return _root.getWsResource();
    }

    public boolean hasNegativeAssertion()
    {
        return _hasNegativeAssertion;
    }

    public boolean matches(EndpointReference epr)
        throws SoapFault
    {
        if (epr == null)
            throw new NullPointerException(_MESSAGES.get("NullRemoteResource"));
        
        return _root.matches(epr);
    }

    public void removeMatch(Match child)
    {
        if (child == null)
            throw new NullPointerException(_MESSAGES.get("NullMatch"));
        
        _root.removeMatch(child);
    }

    public Element toXML()
    {
        return toXML(XmlUtils.EMPTY_DOC);
    }

    public Element toXML(Document doc)
    {
        if (doc == null)
            throw new NullPointerException(_MESSAGES.get("NullDocument"));
        
        QName qname = MuwsConstants.CORRELATABLE_QNAME;
        Element rootXML = XmlUtils.createElement(doc, qname);
        
        rootXML.setAttribute(MuwsConstants.DIALECT, getDialect());
        
        String negative = Boolean.toString(hasNegativeAssertion());
        rootXML.setAttribute(MuwsConstants.NEGATIVE, negative);
        
        Element matchXML = _root.toXML(doc);
        
        //
        // CorrelatableProperties doesn't have one "root" Match, so we 
        // have to remove the pbm:MatchAll element that is containing 
        // all of our Matches
        //
        return (Element)XmlUtils.moveSubTree(matchXML, rootXML);
    }
}
