/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.notification;

import java.util.Date;

import javax.xml.namespace.QName;

import org.w3c.dom.Element;

import org.apache.muse.util.xml.XmlSerializable;
import org.apache.muse.ws.addressing.EndpointReference;
import org.apache.muse.ws.addressing.soap.SoapFault;
import org.apache.muse.ws.notification.faults.*;
import org.apache.muse.ws.notification.topics.Topic;
import org.apache.muse.ws.notification.topics.TopicNamespace;
import org.apache.muse.ws.notification.topics.TopicSet;
import org.apache.muse.ws.notification.topics.WstConstants;
import org.apache.muse.ws.resource.WsResource;
import org.apache.muse.ws.resource.WsResourceCapability;
import org.apache.muse.ws.resource.basefaults.BaseFault;

/**
 * 
 * NotificationProducer is an interface that represents the WS-Notification 
 * NotificationProducer port type; it is based on WS-N 1.3. It provides support 
 * for the NotificationProducer properties (all topic-related) as well as its 
 * two operations: GetCurrentMessage and Subscribe. It also has methods for 
 * managing the resource's topic set using Muse's {@linkplain Topic Topic API}.
 * <br><br>
 * Perhaps the most significant API for users of WS-N is the publish() method; 
 * this method allows users to send message content to subscribers without 
 * having to know anything about the maintenance of subscriptions and how to 
 * determine which messages are sent to which subscribers. The publish() method 
 * allows user code to focus on generating event content and then let the WS-N 
 * implementation determine how and where to put it on the wire.
 * 
 * @author Dan Jemiolo (danj)
 * 
 */

public interface NotificationProducer extends WsResourceCapability
{
    QName[] PROPERTIES = new QName[]{
        WsnConstants.FIXED_SET_QNAME, 
        WsnConstants.TOPIC_DIALECT_QNAME, 
        WsnConstants.TOPIC_EXPRESSION_QNAME, 
        WstConstants.TOPIC_SET_QNAME
    };
    
    /**
     * 
     * This method is for internal (service-side) code that wishes to add a 
     * subscription that has already been created (perhaps from data residing 
     * in a persistent data store) and does not require a call to subscribe(). 
     * It should not be exposed to remote clients - the subscribe() method is 
     * already available for the purpose of remote subscription.
     *
     * @param subscription
     *
     */
    void addSubscription(WsResource subscription);
    
    /**
     * 
     * Creates a topic in the resource's topic set and adds it to the 
     * appropriate topic namespace. If the topic namespace does not exist, 
     * it is created.
     *
     * @param topicName
     *        The name of the topic, where the namespace matches the topic 
     *        namespace the topic belongs to and the local part is a path 
     *        that defines the topic's hierarchy (the path may also be a 
     *        simple name, no slashes).
     *        
     * @return The Topic that was created and added to the resource's topic set.
     * 
     */
    Topic addTopic(QName topicName) 
        throws BaseFault;
    
    /**
     * 
     * Creates a topic namespace in the resource's topic set using the given URI.
     * 
     * @return The TopicNamespace that was created and added to the resource's 
     *         topic set.
     *
     */
    TopicNamespace addTopicNamespace(String namespace)
        throws BaseFault;
    
    /**
     * 
     * Retrieves the last message published for a given topic.
     * <br><br>
     * <b>NOTE:</b>The term 'current' is only helpful if you think of the 
     * published messages being stored in a stack, with the last one on top. 
     * This is the 'current message'. In reality, implementations are not likely 
     * to store an in-memory copy of every message that is published, but 
     * it is a useful mental model.
     *
     * @param topicPath
     *        The name of the Topic whose last message you wish to retrieve.
     * 
     * @return The last message that was published on the given topic.
     * 
     */
    NotificationMessage getCurrentMessage(QName topicPath)
        throws TopicExpressionDialectUnknownFault, 
               InvalidTopicExpressionFault, 
               TopicNotSupportedFault, 
               MultipleTopicsSpecifiedFault, 
               NoCurrentMessageOnTopicFault;
    
    /**
     * 
     * @return True if the resource's topic set cannot be modified <em>after it 
     *         has been initialized</em>.
     *
     */
    boolean getFixedTopicSet()
        throws BaseFault;
    
    /**
     * 
     * @param topicName
     * 
     * @return The Topic with the given namespace and name, or null if no such 
     *         Topic exists.
     *
     */
    Topic getTopic(QName topicName);
    
    /**
     * 
     * @return The names of all of the topic that are supported by this resource.
     *
     */
    QName[] getTopicExpression()
        throws BaseFault;
    
    /**
     * 
     * @return All of the topic expression dialects that are supported by 
     *         this resource.
     *
     */
    String[] getTopicExpressionDialect()
        throws BaseFault;
    
    /**
     * 
     * @param namespace
     * 
     * @return The TopicNamespace representing the given URI, or null if no 
     *         such TopicNamespace exists.
     *
     */
    TopicNamespace getTopicNamespace(String namespace);
    
    /**
     * 
     * @return The resource's entire topic set, which contains topic namespaces 
     *         and (within those) topics.
     *
     */
    TopicSet getTopicSet();
    
    /**
     * 
     * @param topicName
     * 
     * @return True if the resource supports the given topic.
     *
     */
    boolean hasTopic(QName topicName);
    
    /**
     * 
     * Creates a WS-N notification message with the given topic and message 
     * content and sends it to the appropriate subscribers. The implementation 
     * will do all of the work of figuring out which subscribers should get 
     * the message based on its topic and/or content.
     *
     * @param topicName
     *        The WS-N topic associated with the message. This is optional (it 
     *        can be null).
     *        
     * @param content
     *        The XML fragment that will be put under the wsnt:Message section 
     *        of the notification message. This is also called the 'payload'.
     *
     */
    void publish(QName topicName, Element content)
        throws SoapFault;
    
    /**
     * 
     * This is a convenience method that calls publish(QName, Element) after 
     * the XmlSerializable object has been serialized to XML.
     *
     */
    void publish(QName topicName, XmlSerializable content)
        throws SoapFault;
    
    /**
     * 
     * This method is for internal (service-side) code that wishes to remove a 
     * subscription so that messages are no longer published to its subscriber.
     * It should not be exposed to remote clients - the WS-RL capabilities can 
     * be added to the subscription resource type so that subscriptions can be 
     * removed remotely.
     * 
     * @param subscription
     *
     */
    void removeSubscription(EndpointReference subscriptionEPR);
    
    /**
     * 
     * Registers a consumer resource to receive events that are published 
     * to a certain topic (along with other filtering rules). Registration 
     * results in the creation of a subscription resource, which exposes 
     * the WS-N SubscriptionManager port type and can be managed as a regular 
     * WS-RF implied resource. Once subscription is complete, the event consumer 
     * will receive notifications on the topic until the subscription is 
     * terminated (via WS-RL) or the producer shuts down.
     *
     * @param consumer
     *        The EPR of the resource that will receive the notifications 
     *        from the producing resource. This does <b>not</b> have to 
     *        be the same component that is making the subscribe() request. 
     *        An orchestrating service may be subscribing for the consumer.
     *        
     * @param filter
     *        Optional filter that allows for messages to be published if 
     *        they meet a certain criteria. Filter types include topic names, 
     *        message patterns (XPath), and current state (WSRP doc).
     * 
     * @param policy
     *        Optional filter that allows for more complex analysis on the 
     *        conditions needed to send a message. The implementation must 
     *        provide one or more policy languages (dialects) that clients 
     *        can use to express how and when a message that is published 
     *        to a topic should be sent. If this parameter is null, no 
     *        policy is applied.
     * 
     * @param terminationTime
     *        Because subscriptions are WS-resources, they may be terminated 
     *        with scheduled termination (specifically, the WS-RL TerminationTime 
     *        property). If this parameter is null, no termination time is set.
     * 
     * @return The new subscription resource.
     *
     */
    WsResource subscribe(EndpointReference consumer, 
                         Filter filter,  
                         Date terminationTime, 
                         Policy policy)
        throws InvalidFilterFault, 
               TopicExpressionDialectUnknownFault, 
               InvalidTopicExpressionFault, 
               TopicNotSupportedFault, 
               InvalidProducerPropertiesExpressionFault, 
               InvalidMessageContentExpressionFault, 
               UnacceptableInitialTerminationTimeFault, 
               UnrecognizedPolicyRequestFault, 
               UnsupportedPolicyRequestFault, 
               NotifyMessageNotSupportedFault, 
               SubscribeCreationFailedFault;
}
