/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.notification;

import java.util.Date;

import javax.xml.namespace.QName;

import org.apache.muse.util.xml.XmlSerializable;
import org.apache.muse.ws.addressing.EndpointReference;
import org.apache.muse.ws.addressing.soap.SoapFault;
import org.apache.muse.ws.notification.faults.PauseFailedFault;
import org.apache.muse.ws.notification.faults.ResumeFailedFault;
import org.apache.muse.ws.resource.WsResourceCapability;

/**
 * 
 * SubscriptionManager is an interface that represents the WS-Notification 
 * SubscriptionManager port type; it is based on WS-N v1.3. Resources that 
 * use this capability represent an agreement between a resource and the 
 * consumers that are listening to its notifications. Destroying a subscription 
 * resource ends the relationship and the consumer will no longer receive 
 * messages from the producer resource.
 * <br><br>
 * Muse represents all WS-N subscriptions at WS-RF implied resources. They 
 * are accessible through a unique endpoint and can be destroyed via WS-RL 
 * if the WS-RL capabilities are used. Adding this capability to a resource 
 * type gives it all of the required features of a WS-N subscription resource.
 * 
 * @author Dan Jemiolo (danj)
 *
 */

public interface SubscriptionManager extends WsResourceCapability, XmlSerializable
{
    QName[] PROPERTIES = new QName[]{
        WsnConstants.CONSUMER_QNAME, 
        WsnConstants.FILTER_QNAME, 
        WsnConstants.POLICY_QNAME, 
        WsnConstants.CREATION_QNAME 
    };
    
    /**
     * 
     * @return The EPR of the resource that is receiving the notifications.
     *
     */
    EndpointReference getConsumerReference();
    
    /**
     * 
     * @return The time that the subscription resource was created.
     *
     */
    Date getCreationTime();
    
    /**
     * 
     * @return The filter that provided when the subscription was created.
     *
     */
    Filter getFilter();
    
    /**
     * 
     * @return The EPR of the resource that created the subscription (when 
     *         subscribe() was called).
     *
     */
    EndpointReference getProducerReference();

    /**
     * 
     * @return The policy that provided when the subscription was created.
     *
     */
    Policy getSubscriptionPolicy();
    
    /**
     * 
     * @return True if the subscription is paused and not publishing messages 
     *         to its consumer.
     *
     */
    boolean isPaused();

    /**
     * 
     * Temporarily stops the producer resource from sending messages that 
     * are published to consumer. This does not involved terminating the 
     * subscription with WS-RL - the subscription resource will remain 
     * intact through the pause. Messages will only be sent again if the 
     * resumeSubscription() operation is invoked on the subscription.
     * <br><br>
     * If the subscription is already paused, the result is a no-op (the 
     * operation is idempotent).
     *
     */
    void pauseSubscription()
        throws PauseFailedFault;
    
    /**
     * 
     * Sends the given message to the subscription's consumer resource.
     * 
     * @param message
     *
     */
    void publish(NotificationMessage message)
        throws SoapFault;
    
    /**
     * 
     * If the subscription is paused, it is unpaused, so that messages 
     * published to the topic will once again be sent to the consumer. 
     * If the subscription was not paused, the result is a no-op (the 
     * operation is idempotent).
     *
     */
    void resumeSubscription()
        throws ResumeFailedFault;
    
    void setConsumerReference(EndpointReference consumer);
    
    void setFilter(Filter filter);
    
    void setProducerReference(EndpointReference producer);
    
    void setSubscriptionPolicy(Policy policy);
}
