/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.notification.impl;

import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XPathUtils;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.notification.Filter;
import org.apache.muse.ws.notification.NotificationMessage;
import org.apache.muse.ws.notification.WsnConstants;
import org.apache.muse.ws.notification.faults.InvalidMessageContentExpressionFault;

/**
 * 
 * MessagePatternFilter is the implementation of message pattern filters for 
 * WS-N subscriptions. The dialect for message patterns is XPath 1.0.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public class MessagePatternFilter implements Filter
{
    //
    // Used to look up all exception messages
    //
    private static Messages _MESSAGES = MessagesFactory.get(MessagePatternFilter.class);
    
    //
    // the xpath that will be evaluated against outgoing messages
    //
    private String _pattern = null;
    
    public MessagePatternFilter(String pattern, String dialect) 
        throws InvalidMessageContentExpressionFault
    {
        //
        // we only support xpath
        //
        if (!dialect.equals(XPathUtils.NAMESPACE_URI))
        {
            Object[] filler = { dialect, XPathUtils.NAMESPACE_URI };
            throw new InvalidMessageContentExpressionFault(_MESSAGES.get("InvalidDialect", filler));
        }
        
        _pattern = pattern;
    }
    
    public boolean accepts(NotificationMessage message)
    {
        //
        // we must append the message XML to the document so we can 
        // can resolve /absolute XPaths in addition to relative ones
        //
        Document doc = XmlUtils.createDocument();
        Element messageXML = message.toXML(doc);
        doc.appendChild(messageXML);
        
        try
        {
            return XPathUtils.isMatch(messageXML, _pattern);
        }
        
        catch (TransformerException error)
        {
            return false;
        }
    }
    
    public String toString()
    {
        return XmlUtils.toString(toXML(), false);
    }

    public Element toXML()
    {
        return toXML(XmlUtils.EMPTY_DOC);
    }

    public Element toXML(Document doc)
    {
        Element filter = XmlUtils.createElement(doc, WsnConstants.FILTER_QNAME);
        
        Element message = XmlUtils.createElement(doc, WsnConstants.MESSAGE_CONTENT_QNAME, _pattern);
        message.setAttribute(WsnConstants.DIALECT, XPathUtils.NAMESPACE_URI);
        filter.appendChild(message);
        
        return filter;
    }
}
