/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.notification.impl;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.muse.util.LoggingUtils;
import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XPathUtils;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.notification.Filter;
import org.apache.muse.ws.notification.NotificationMessage;
import org.apache.muse.ws.notification.WsnConstants;
import org.apache.muse.ws.notification.faults.*;
import org.apache.muse.ws.resource.WsResource;
import org.apache.muse.ws.resource.properties.ResourcePropertyCollection;

/**
 * 
 * ProducerPropertiesFilter is the implementation of resource property filters 
 * for WS-N subscriptions. The filter objects need a reference to the producer 
 * resource so that they can read current property values each time the filter 
 * is evaluated. This filter is different from others in that it doesn't 
 * evaluate the message content; rather, it evalutes the resource's current 
 * WS-RP document in order to decide if it should send the message.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public class ProducerPropertiesFilter implements Filter
{
    //
    // Used to look up all exception messages
    //
    private static Messages _MESSAGES = MessagesFactory.get(ProducerPropertiesFilter.class);
    
    //
    // the resource whose ws-rp doc is being evaluated by the filter
    //
    private WsResource _resource = null;
    
    //
    // xpath expression to evaluate on the ws-rp doc
    //
    private String _expression = null;
    
    public ProducerPropertiesFilter(WsResource resource, 
                                    String expression, 
                                    String dialect) 
        throws InvalidProducerPropertiesExpressionFault
    {
        //
        // we only support xpath
        //
        if (!dialect.equals(XPathUtils.NAMESPACE_URI))
        {
            Object[] filler = { dialect, XPathUtils.NAMESPACE_URI };
            throw new InvalidProducerPropertiesExpressionFault(_MESSAGES.get("InvalidDialect", filler));
        }
        _resource = resource;
        _expression = expression;
    }
    
    public boolean accepts(NotificationMessage message)
    {
        //
        // evaluate the xpath over the resource's current ws-rp doc - if the 
        // result is 'true', we publish the message
        //
        try
        {
            ResourcePropertyCollection props = _resource.getPropertyCollection();
            Element wsrpDoc = props.getResourcePropertyDocument();
            return XPathUtils.isMatch(wsrpDoc, _expression);
        }
        
        catch (Throwable error)
        {
            LoggingUtils.logError(_resource.getLog(), error);
            return false;
        }
    }
    
    public String toString()
    {
        return XmlUtils.toString(toXML(), false);
    }

    public Element toXML()
    {
        return toXML(XmlUtils.EMPTY_DOC);
    }

    public Element toXML(Document doc)
    {
        Element filter = XmlUtils.createElement(doc, WsnConstants.FILTER_QNAME);
        
        Element props = XmlUtils.createElement(doc, WsnConstants.PRODUCER_PROPERTIES_QNAME, _expression);
        props.setAttribute(WsnConstants.DIALECT, XPathUtils.NAMESPACE_URI);
        filter.appendChild(props);
        
        return filter;
    }
}
