/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.notification.properties;

import javax.xml.namespace.QName;

import org.w3c.dom.Element;

import org.apache.muse.ws.addressing.soap.SoapFault;
import org.apache.muse.ws.notification.NotificationProducer;
import org.apache.muse.ws.resource.basefaults.BaseFault;
import org.apache.muse.ws.resource.basefaults.WsbfUtils;
import org.apache.muse.ws.resource.properties.listeners.PropertyChangeListener;

/**
 *
 * ChangeNotificationListener is a {@linkplain PropertyChangeListener PropertyChangeListener} 
 * that reports changes in a resource property to the WS-N topic that represents 
 * the property. The WS-N notification message create by this listener have a 
 * <em>ResourcePropertyValueChangeNotification</em> content item. That content 
 * contains the old and new values of the property.
 * <br><br>
 * Note that this class does not send the notification messages directly - that 
 * is handled by the WS-N NotificationProducer component.
 *
 * @author Dan Jemiolo (danj)
 * 
 * @see WsrpNotificationListenerFactory
 *
 */

public class ChangeNotificationListener implements PropertyChangeListener
{
    //
    // The name of the property (topic) we're reporting on
    //
    private QName _topicName = null;
    
    private NotificationProducer _wsn = null;
    
    public ChangeNotificationListener(QName topicName, NotificationProducer wsn)
    {
        _topicName = topicName;
        _wsn = wsn;
    }
    
    protected NotificationProducer getNotificationProducer()
    {
        return _wsn;
    }
    
    public QName getPropertyName()
    {
        return _topicName;
    }
    
    /**
     * 
     * Creates a WS-N message that holds a WS-RP change notification event. 
     * The format of this event is the one recommended by the WS-RP spec for 
     * integrating with WS-N; it has a root element named 
     * <em>ResourcePropertyValueChangeNotification</em> that wraps two child 
     * elements: the old value and the new value. Once the message is created, 
     * it is set as the current message for the property's WS-N SimpleTopic.
     * <br><br>
     * This method does not actually send out any notifications. The WS-N 
     * framework will send out the message in the proper format if there are 
     * any subscribers to the property's topic at the time of the change.
     *
     */
    public void propertyChanged(Element oldValue, Element newValue)
        throws BaseFault
    {
        QName propertyName = getPropertyName();
        Element[] oldValues = new Element[]{ oldValue };
        Element[] newValues = new Element[]{ newValue };
        
        ResourcePropertyChangeEvent event = new ResourcePropertyChangeEvent(propertyName, oldValues, newValues);
        
        //
        // publish to the property's topic
        //
        try
        {
            getNotificationProducer().publish(propertyName, event);
        }
        
        catch (SoapFault error)
        {
            throw WsbfUtils.convertToFault(error);
        }
    }
}
