/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.resource.lifetime;

import java.util.Date;

import javax.xml.namespace.QName;

import org.apache.muse.ws.resource.WsResourceCapability;
import org.apache.muse.ws.resource.basefaults.BaseFault;
import org.apache.muse.ws.resource.lifetime.faults.TerminationTimeChangeRejectedFault;
import org.apache.muse.ws.resource.lifetime.faults.UnableToSetTerminationTimeFault;

/**
 * 
 * ScheduledTermination represents the WS-ResourceLifetime ScheduledResourceTermination 
 * capability, which has two properties (wsrf-rl:CurrentTime and 
 * wsrf-rl:TerminationTime) and one method (wsrf-rl:SetTerminationTime). 
 * Implementations of this capability provide the self-destruction feature 
 * that occurs when the system's current time reaches the resource's 
 * termination time.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public interface ScheduledTermination extends WsResourceCapability
{
    QName[] PROPERTIES = new QName[]{
        WsrlConstants.CURRENT_TIME_QNAME, 
        WsrlConstants.TERMINATION_TIME_QNAME    
    };
    
    /**
     * 
     * @return The current time, according to the system the resource is on.
     *
     */
    Date getCurrentTime() 
        throws BaseFault;
    
    /**
     * 
     * @return The termination time for this resource, or null if the resource 
     *         is not scheduled for termination.
     *
     */
    Date getTerminationTime() 
        throws BaseFault;
    
    /**
     * 
     * @param time
     *        The new termination time. If the time is null, the resource 
     *        will not be scheduled for destruction. If the time is in the 
     *        past, it will be destroyed immediately. 
     * 
     * @return The previous termination time value (or null, if none existed).
     *  
     * @throws UnableToSetTerminationTimeFault
     *         <ul>
     *         <li>If the new time is invalid or the current time cannot 
     *         be adjusted.</li>
     *         </ul>
     *         
     * @throws TerminationTimeChangeRejectedFault
     *         <ul>
     *         <li>If the new time is invalid according to a policy of 
     *         the WS-RL implementation.</li>
     *         </ul>
     *
     */
    Date setTerminationTime(Date time) 
        throws UnableToSetTerminationTimeFault, 
               TerminationTimeChangeRejectedFault;
}
