/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.resource.sg;

import java.util.Date;

import javax.xml.namespace.QName;

import org.w3c.dom.Element;

import org.apache.muse.ws.addressing.EndpointReference;
import org.apache.muse.ws.resource.WsResource;
import org.apache.muse.ws.resource.WsResourceCapability;
import org.apache.muse.ws.resource.basefaults.BaseFault;
import org.apache.muse.ws.resource.sg.MembershipContentRule;
import org.apache.muse.ws.resource.sg.faults.AddRefusedFault;
import org.apache.muse.ws.resource.sg.faults.ContentCreationFailedFault;
import org.apache.muse.ws.resource.sg.faults.UnsupportedMemberInterfaceFault;

/**
 * 
 * ServiceGroup is an internal representation of the WSDM ServiceGroup 
 * capability. It provides access to the service group properties along 
 * with methods for creating, destroying, and querying entries. Resources 
 * that use this capability will have the ability to act as directories for 
 * other resources while still performing other functions.
 * 
 * @author Dan Jemiolo (danj)
 * 
 */

public interface ServiceGroup extends WsResourceCapability
{
    QName[] PROPERTIES = new QName[]{
        WssgConstants.ENTRY_QNAME, 
        WssgConstants.CONTENT_RULE_QNAME, 
    };
    
    /**
     * 
     * Creates a new Entry resource and adds it to Muse's ResourceManager and 
     * the ServiceGroup's WS-RP container.
     * 
     * @param memberEPR
     *        The EPR of the resource that the Entry will represent.
     * 
     * @param termination
     *        The initial termination time for the Entry. This may be null 
     *        if the Entry is not scheduled for destruction.
     * 
     * @return The newly-created (and stored) Entry resource.
     *
     */
    WsResource addEntry(EndpointReference memberEPR, 
                        Element content, 
                        Date termination)
        throws AddRefusedFault, 
               ContentCreationFailedFault, 
               UnsupportedMemberInterfaceFault, 
               BaseFault;
    
    /**
     * 
     * Adds the EPR/resource mapping to the implementation's collection 
     * of service group entries.
     * 
     * @param memberEPR
     *        The EPR of the service group entry.
     * 
     * @param entry
     *        The resource that represents the service group entry.
     * 
     * @return The same resource provided as a second parameter.
     *
     */
    WsResource addEntry(EndpointReference memberEPR, WsResource entry)
        throws AddRefusedFault, 
               ContentCreationFailedFault, 
               UnsupportedMemberInterfaceFault, 
               BaseFault;
    
    /**
     * 
     * @return All of the content elements in the group's membership content rules.
     *
     */
    QName[] getContentElements()
        throws BaseFault;
    
    WsResource[] getEntry()
        throws BaseFault;
    
    /**
     * 
     * @param memberEPR
     * 
     * @return The entry resource associated with the given EPR, or null if 
     *         no such resource exists.
     *
     */
    WsResource getEntry(EndpointReference memberEPR)
        throws BaseFault;
    
    MembershipContentRule[] getMembershipContentRule()
        throws BaseFault;
    
    /**
     * 
     * @param epr
     *        The EPR for the resource against which all the service group's 
     *        membership content rules should be evaluated.
     * 
     * @return True if the given resource's WS-RP document matches all of 
     *         the service group's membership rules and can be safely added 
     *         to the service group.
     *
     */
    boolean isMatch(EndpointReference epr)
        throws BaseFault;
    
    void removeEntry(WsResource entry)
        throws BaseFault;
    
    void setMembershipContentRule(MembershipContentRule[] mcr)
        throws BaseFault;
}
