/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.resource.metadata.impl;

import javax.xml.namespace.QName;

import org.w3c.dom.Element;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.resource.basefaults.BaseFault;
import org.apache.muse.ws.resource.metadata.MetadataDescriptor;
import org.apache.muse.ws.resource.properties.listeners.AbstractChangeApprover;
import org.apache.muse.ws.resource.properties.listeners.PropertyChangeApprover;
import org.apache.muse.ws.resource.properties.set.faults.UnableToModifyResourcePropertyFault;

/**
 *
 * ValidValuesApprover is a 
 * {@linkplain PropertyChangeApprover PropertyChangeApprover} that throws 
 * an exception if a caller tries to insert or update a property instance 
 * with a value that is not defined in the WS-RMD document's ValidValues 
 * section. If a property's metadata defines valid values, those are the only 
 * values it may have; if not, any value is acceptable.
 * 
 * @author Dan Jemiolo (danj)
 *
 */

public class ValidValuesApprover extends AbstractChangeApprover
{
    //
    // Used to lookup all exception messages
    //
    private static Messages _MESSAGES = 
        MessagesFactory.get(ValidValuesApprover.class);
    
    //
    // The RMD with the valid values of the properties
    //
    private MetadataDescriptor _metadata = null;
    
    /**
     * 
     * Creates a new approver for the property with the given name and RMD.
     *
     * @param qname
     *        The QName of the property to monitor.
     * 
     * @param metadata
     *        The RMD descriptor with the metadata for the property.
     *
     */
    public ValidValuesApprover(QName qname, MetadataDescriptor metadata)
    {
        super(qname);

        if (metadata == null)
            throw new NullPointerException(_MESSAGES.get("NullMetadataDescriptor"));
        
        _metadata = metadata;
    }
    
    /**
     * 
     * Throws an exception if the the second parameter is not null (signifying 
     * an insertion or update) and is not one of the valid values defined in 
     * the RMD.
     *
     */
    public void validateChange(Element oldValue, Element newValue, Object token)
        throws BaseFault
    {
        QName qname = getPropertyName();
        
        //
        // assuming it's not a deletion (null), check the value. if there 
        // are no ValidValues defined, this should still work according to 
        // the MetadataDescriptor interface.
        //
        if (newValue != null && !_metadata.isValidValue(qname, newValue))
        {
            Object[] filler = { qname, XmlUtils.toString(newValue, false) };
            throw new UnableToModifyResourcePropertyFault(_MESSAGES.get("InvalidValue", filler));
        }
    }
}
