/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.resource.properties.schema.impl;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.w3c.dom.Element;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.util.xml.XsdUtils;
import org.apache.muse.ws.resource.WsResourceCapability;
import org.apache.muse.ws.resource.properties.WsrpConstants;
import org.apache.muse.ws.resource.properties.schema.ResourcePropertiesSchema;
import org.apache.muse.ws.wsdl.WsdlUtils;

/**
 *
 * SimpleResourcePropertiesSchema is Muse's default implementation of the 
 * ({@linkplain ResourcePropertiesSchema WS-RP document schema}). Instances of 
 * this schema structure can be created from a DOM Element containing the WS-RP 
 * document definition.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public class SimpleResourcePropertiesSchema implements ResourcePropertiesSchema
{
    //
    // Used to lookup all exception messages
    //
    private static Messages _MESSAGES = 
        MessagesFactory.get(SimpleResourcePropertiesSchema.class);
    
    //
    // All of the type definitions in the schema. This is a 
    // Map[QName, PropertySchemaDefinition].
    //
    private Map _definitionsByQName = new HashMap();
    
    //
    // The name of the root element in the WSRP doc - this should be 
    // whatever is used in the WSDL, but a default is provided
    //
    private QName _wsrpName = WsrpConstants.DEFAULT_DOCUMENT_QNAME;
    
    /**
     * 
     * Creates a new schema by parsing the WS-RP document and storing the 
     * various property data for future lookups. The child xsd:elements in 
     * the given Element represent the actual properties in the WS-RP 
     * definition; the types of these properties must be referred to using 
     * the XSD <em>ref</em> attribute.
     *
     * @param wsrpElement
     *        The DOM Element containing the WS-RP document's type definition.
     *
     */
    public SimpleResourcePropertiesSchema(QName wsrpName, Element wsrpElement)
    {
        if (wsrpName != null)
            _wsrpName = wsrpName;
        
        if (wsrpElement == null)
            throw new NullPointerException(_MESSAGES.get("NullPropertiesElement"));
        
        QName qname = XmlUtils.getElementQName(wsrpElement);
        
        //
        // make sure we're starting from an element fragment
        //
        if (!qname.equals(XsdUtils.ELEMENT_QNAME))
            throw new RuntimeException(_MESSAGES.get("NotAnElement"));
        
        //
        // get the name of the WS-RP collection
        //
        String name = wsrpElement.getAttribute(XsdUtils.NAME);
        
        if (name == null || name.length() == 0)
            throw new RuntimeException(_MESSAGES.get("NoPropertiesName"));
        
        qname = XmlUtils.parseQName(name, wsrpElement);
        
        //
        // get all of the elements defined in the WS-RP sequence
        //
        // NOTE: we only get those properties declared with "ref"
        //
        Element[] elements = getAllPropertyElements(wsrpElement);        
        
        //
        // for every property element, record its metadata/definition
        //
        for (int n = 0; n < elements.length; ++n)
        {
            String ref = elements[n].getAttribute(XsdUtils.REF);
            qname = XmlUtils.parseQName(ref, elements[n]);
            
            PropertySchemaDefinition definition = 
                new PropertySchemaDefinition(qname, elements[n]);
            _definitionsByQName.put(qname, definition); 
        }
    }
    
    /**
     *
     * @param wsrpElement
     *        The DOM Element containing the WS-RP document definition.
     * 
     * @return All of the child xsd:element Elements that have a <em>ref</em> 
     *         attribute { complexType/sequence/element[@ref] }, including 
     *         those in the schema's base types.
     *
     */
    private Element[] getAllPropertyElements(Element wsrpElement)
    {
        Element complexType = XmlUtils.getElement(wsrpElement, XsdUtils.COMPLEX_TYPE_QNAME);
        
        //
        // if the complexType isn't under the element, there must be a type 
        // attribute that tells us where the definition is
        //
        if (complexType == null)
        {
            Element wsdl = XmlUtils.getDocumentRoot(wsrpElement);
            
            String typeName = wsrpElement.getAttribute(XsdUtils.TYPE);
            QName typeQName = XmlUtils.parseQName(typeName, wsrpElement);
            
            //
            // search the <wsdl:types/> for the WSRP type def
            //
            complexType = WsdlUtils.getComplexTypeDeclaration(wsdl, typeQName);
        }
        
        Element sequence = XmlUtils.getElement(complexType, XsdUtils.SEQUENCE_QNAME);
        
        if (sequence != null)
            return XmlUtils.getElements(sequence, XsdUtils.ELEMENT_QNAME);
        
        return getBasePropertyElements(complexType);
    }
    
    private Element[] getBasePropertyElements(Element complexType)
    {
        Element sequence = XmlUtils.getElement(complexType, XsdUtils.SEQUENCE_QNAME);
        
        //
        // complexType has no parent type - just return its elements
        //
        if (sequence != null)
            return XmlUtils.getElements(sequence, XsdUtils.ELEMENT_QNAME);
        
        Element wsdl = XmlUtils.getDocumentRoot(complexType);
        
        Element complexContent = XmlUtils.getElement(complexType, XsdUtils.COMPLEX_CONTENT_QNAME);
        Element extension = XmlUtils.getElement(complexContent, XsdUtils.EXTENSION_QNAME);
        
        //
        // otherwise, get the parent type and its elements...
        //
        String base = extension.getAttribute(XsdUtils.BASE);
        QName baseName = XmlUtils.parseQName(base, extension);
        Element baseType = WsdlUtils.getComplexTypeDeclaration(wsdl, baseName);
        Element[] baseProperties = getBasePropertyElements(baseType);
        
        //
        // ...now get our own elements...
        //
        sequence = XmlUtils.getElement(extension, XsdUtils.SEQUENCE_QNAME);
        Element[] childProperties = XmlUtils.getElements(sequence, XsdUtils.ELEMENT_QNAME);
        
        //
        // ...and combine the two into one big array
        //
        Element[] allProperties = new Element[baseProperties.length + childProperties.length];
        System.arraycopy(baseProperties, 0, allProperties, 0, baseProperties.length);
        System.arraycopy(childProperties, 0, allProperties, baseProperties.length, childProperties.length);
        
        return allProperties;
    }
    
    public WsResourceCapability getCapability(QName property)
    {
        return getProperty(property).getCapability();
    }
    
    public synchronized final QName getElementName()
    {
        return _wsrpName;
    }

    public int getMaxOccurs(QName property)
    {
        return getProperty(property).getMaxOccurs();
    }
    
    public int getMinOccurs(QName property)
    {
        return getProperty(property).getMinOccurs();
    }

    protected PropertySchemaDefinition getProperty(QName property)
    {
        PropertySchemaDefinition definition = 
            (PropertySchemaDefinition)_definitionsByQName.get(property);

        if (definition == null)
        {
            Object[] filler = { property };
            throw new RuntimeException(_MESSAGES.get("PropertyNotFound", filler));
        }
        
        return definition;        
    }
    
    public Collection getPropertyNames()
    {
        return Collections.unmodifiableSet(_definitionsByQName.keySet());
    }
    
    public QName getPropertyTypeName(QName property)
    {
        return getProperty(property).getPropertyTypeName();
    }
    
    public boolean hasCapability(QName property)
    {
        return getCapability(property) != null;
    }
    
    public boolean hasProperty(QName property)
    {
        return _definitionsByQName.containsKey(property);
    }

    public boolean isMaxUnbounded(QName property)
    {
        return getProperty(property).isMaxUnbounded();
    }

    public boolean isNillable(QName property)
    {
        return getProperty(property).isNillable();
    }

    public void setCapability(QName property, WsResourceCapability capability)
    {
        getProperty(property).setCapability(capability);
    }
    
    public synchronized void setElementName(QName wsrpName)
    {
        if (wsrpName == null)
            throw new NullPointerException(_MESSAGES.get("NullWSRPElementName"));
        
        _wsrpName = wsrpName;
    }
}
